"""
    linkedin-api-no-cookie
"""
from .linkedin import Linkedin

__title__ = "linkedin_api_no_cookie"
__version__ = "1.12"
__description__ = "Python Wrapper for the Linkedin API"

__license__ = "MIT"

__author__ = "Alejandrina Gonzalez"
__email__ = "me@alejandrina.me"

__all__ = ["Linkedin"]
